from typing import Dict, Optional, Any
from datetime import datetime as dt
import logging

import requests
from faker import Faker

import ckan.plugins.toolkit as tk
import ckan.logic as logic

import ckanext.msal.config as msal_conf
import ckanext.msal.utils as msal_utils


log = logging.getLogger(__name__)
USER_ENDPOINT = "https://graph.microsoft.com/beta/me"


def _login_user(user_data: dict) -> Dict[str, Any]:
    return get_or_create_user(user_data)


def get_or_create_user(user_data: Dict[str, Any]) -> Dict[str, Any]:
    """Returns an existed user data by oid (object id) or
    creates a new one from user_data fetched from Microsoft Graph API

    args
    user_data: dict - user data dict

    {
        "displayName": "Mark Spencer",
        "givenName": "Mark",
        "mailNickname": "mark209",
        "mail": mark209@myorg.onmicrosoft.com,
        "surname": "Spencer",
        "userPrincipalName": "mark209@myorg.onmicrosoft.com",
        "id": "3f22cb88-c272-44f1-838f-f823cdc08bd6",
    }

    return
    type: Dict[str, Any]
    """

    object_id: str = user_data["oid"]

    try:
        user: Dict[str, Any] = tk.get_action("user_show")(
            {"ignore_auth": True}, {"id": object_id}
        )
    except logic.NotFound:
        log.info(f"MSAL. User not found, creating new one.")
        user_dict = get_msal_user_data()

        if "error" in user_dict:
            log.error("MSAL. User creation failed.")
            return {}
        return _create_user_from_user_data(user_dict, object_id) if user_dict else {}

    return user


def get_msal_user_data() -> Dict[str, Any]:
    """
    Requests an additional user data from microsoft graph API

    return
    type: Dict[str, Any]
    """
    token: Optional[Dict[Any, Any]] = msal_utils._get_token_from_cache(msal_conf.SCOPE)
    if not token:
        return {}

    log.info(f"MSAL. Fetching user data from Microsoft Graph API by an access token.")
    resp = requests.get(
        USER_ENDPOINT,
        headers={"Authorization": "Bearer " + token["access_token"]},
        params={
            "$id"
            "$select": "id,displayName,userPrincipalName,mail,mailNickname,accountEnabled"
        },
    )

    try:
        resp.raise_for_status()
    except requests.HTTPError:
        log.error(f"MSAL. Fetch failed: {resp.reason} {resp.status_code}")
        return resp.json()

    log.info(f"MSAL. Success fetch.")

    user_data: dict[str, Any] = resp.json()
    user_email: str = _get_email(user_data)

    if (msal_utils.is_email_restricted(user_email)
        or not msal_utils.is_email_allowed(user_email)):
        log.info(
            "MSAL. User won't be created, "
            f"because of the domain policy: {user_email}"
        )
        return {
            "error": tk._(msal_conf.RESTRICTION_ERR)
        }
    return resp.json()


def _create_user_from_user_data(user_data: dict, object_id: str) -> Dict[str, Any]:
    """Create a user with random password using Microsoft Graph API's data.

    raises
    ValidationError if email is not unique

    args
    user_data: dict - user data dict
    object_id: str - actually a `user_id` from Azure AD

    return
    type: dict
    """
    _id: str = object_id
    email: str = _get_email(user_data)
    password: str = msal_utils._make_password()
    username: str = _get_username(user_data)

    if not _is_username_unique(username):
        username = f"{username}-{dt.now().strftime('%S%f')}"

    user = tk.get_action("user_create")(
        {"ignore_auth": True, "user": username},
        {
            "id": _id,
            "email": email,
            "name": username,
            "password": password,
        },
    )
    log.info(f"MSAL. User has been created: {user['id']} - {user['name']}.")
    return user


def _get_email(user_dict: Dict[str,str,]) -> str:
    """
    Fetches email from user_data if exists, otherwise generates random email
    The `userPrincipalName` is formatted like an email address (username@onmicrosoft.com)
    """
    return user_dict.get("userPrincipalName") or user_dict.get("mail") or _make_email()


def _make_email(domain: str = "onmicrosoft.com") -> str:
    """
    Returns a random email with custom domain
    If domain is not provided uses `onmicrosoft.com`

    args
    domain: str - domain used to generate email

    return
    type: str
    """
    f = Faker()
    return f.email()


def _get_username(user_dict: Dict[str, str]) -> str:
    """
    Fetches username from user_data if exists
    If not - munges it from userPrincipalName or mail

    args
    user_dict: Dict[str, str] - user data dict

    return
    type: str
    """
    username: Optional[str] = user_dict.get("mailNickname")

    if not username:
        username = user_dict.get("mail") or user_dict["userPrincipalName"]
        return username.split("@")[0]
    return username


def _is_username_unique(username: str) -> bool:
    try:
        tk.get_action("user_show")({"ignore_auth": True}, {"id": username})
    except logic.NotFound:
        return True

    return False


def is_user_enabled(user_dict: Dict[str, str]) -> bool:
    """
    Returns True if user is enabled

    # TODO
    Currently, we are not using this function.

    args
    user_dict: Dict[str, str] - user data dict

    return
    type: bool
    """
    return user_dict.get("accountEnabled") is True


def is_user_sysadmin(user_dict: Dict[str, str]) -> bool:
    """
    Returns True if user is sysadmin

    # TODO
    Currently, we are not using this function.
    I didn't find an apropriate property to say if user is a sysadmin yet

    args
    user_dict: Dict[str, str] - user data dict

    return
    type: bool
    """
    return False



from datetime import datetime as dt

from flask import Blueprint

import ckan.lib.helpers as h
import ckan.plugins.toolkit as tk
from ckan.common import session, request

import ckanext.msal.config as msal_conf
import ckanext.msal.utils as msal_utils
from ckanext.msal.user import get_msal_user_data



msal = Blueprint('msal', __name__)


@msal.route(msal_conf.REDIRECT_PATH)
def authorized():
    try:
        cache = msal_utils._load_cache()
        result = msal_utils.build_msal_app(cache=cache).acquire_token_by_auth_code_flow(
            session.get("msal_auth_flow", {}), request.args)

        if "error" in result:
            session.clear()

            h.flash_error(
                tk._("Login error. Contact administrator."))
            return h.redirect_to(h.url_for("user.login"))

        session["user"] = result.get("id_token_claims")
        session["user_exp"] = msal_utils._get_exp_date()
        msal_utils._save_cache(cache)

        user_data = get_msal_user_data()
        
        if user_data.get('error'):
            session.clear()
            h.flash_error(user_data["error"])
            return h.redirect_to("user.login")
        
    except ValueError:
        # Usually caused by CSRF
        # Simply ignore them
        pass
    
    return h.redirect_to(h.url_for("dashboard.index"))


@msal.route("/user/msal-logout")
def logout():
    session.clear()  # Wipe out user and its token cache from session
    return h.redirect_to(
            f"{msal_conf.AUTHORITY}/oauth2/v2.0/logout?post_logout_redirect_uri={h.url_for('user.logout', _external=True)}")


@msal.route("/user/msal-login")
def login():
    flow = msal_utils.build_auth_code_flow(scopes=msal_conf.SCOPE)
    session["msal_auth_flow"] = flow

    return h.redirect_to(flow["auth_uri"], _external=True)


def get_blueprints():
    return [msal]



import ckan.plugins as p
import ckan.plugins.toolkit as tk
import ckan.model as model
from ckan.common import session

from ckanext.msal.middleware import SessionInvalidator
from ckanext.msal.views import get_blueprints
import ckanext.msal.user as user_funcs


class MsalPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurer)
    p.implements(p.IAuthenticator, inherit=True)
    p.implements(p.IMiddleware, inherit=True)
    p.implements(p.IBlueprint, inherit=True)

    # IConfigurer

    def update_config(self, config_):
        tk.add_template_directory(config_, 'templates')

    # IMiddleware
    def make_middleware(self, app, config):
        return SessionInvalidator(app)

    # IBlueprint
    def get_blueprint(self):
        return get_blueprints()

    # IAuthenticator
    def identify(self):
        u'''Called to identify the user.

        If the user is identified then it should set:

         - tk.g.user: The name of the user
         - tk.g.userobj: The actual user object
        '''

        if session.get('user') and not any((tk.g.setdefault('userobj'), tk.g.setdefault('user'))):
            user = user_funcs._login_user(session['user'])

            if user:
                tk.g.user = user["name"]
            else:
                session.clear()
    
    def logout(self):
        session.clear()


import pytest
import mock

from ckan.common import config, session

from ckanext.msal.user import _login_user


@pytest.mark.ckan_config("ckan.plugins", "msal")
@pytest.mark.usefixtures("with_plugins", "with_request_context")
def test_login(app):
    user = {
        "aud": "45f7852c-628c-437f-b3f2-2df26976b1e5",
        "iss": "https://login.microsoftonline.com/3c8827a9-65fe-40b5-8644-3173d7026601/v2.0",
        "iat": 1630936420,
        "nbf": 1630936420,
        "exp": 1630940320,
        "name": "Mark Spencer",
        "nonce": "a8d30f17e256f8d5ae042cca2087e8a848871d080ad4349248fce49c558ba7cb",
        "oid": "fb9c93ba-0768-4816-8fcc-802b588fb8bf",
        "preferred_username": "kvaqich@kvaqich.onmicrosoft.com",
        "rh": "0.AQwAqSeIPP5ltUCGRDFz1wJmASyF90WMYn9Ds_It8ml2seUMAHo.",
        "sub": "uXUZ-tk_v6B1Ix82gyQhyPp8CR_CLu5MUVmtkq4Y0To",
        "tid": "3c8827a9-65fe-40b5-8644-3173d7026601",
        "uti": "acThZD9K4EaUOYdVYkxaAA",
        "ver": "2.0",
    }
    with mock.patch(
        "ckanext.msal.user.get_msal_user_data",
        return_value={
            "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users/$entity",
            "@odata.id": "https://graph.microsoft.com/v2/3c8827a9-65fe-40b5-8644-3173d7026601/directoryObjects/fb9c93ba-0768-4816-8fcc-802b588fb8bf/Microsoft.DirectoryServices.User",
            "businessPhones": ["380999222611"],
            "displayName": "Mark Spencer",
            "givenName": "Mark",
            "mailNickname": "mark209",
            "jobTitle": None,
            "mail": None,
            "mobilePhone": None,
            "officeLocation": None,
            "preferredLanguage": None,
            "surname": "Spencer",
            "userPrincipalName": "kvaqich@kvaqich.onmicrosoft.com",
            "id": "fb9c93ba-0768-4816-8fcc-802b588fb8bf",
        },
    ) as get_msal_user_data:
        user_data = _login_user(user)
        assert user["oid"] == user_data["id"]



from ckan.common import config


CLIENT_ID = config.get("ckanext.msal.client_id")
CLIENT_SECRET = config.get("ckanext.msal.client_secret")
AUTHORITY = f"https://login.microsoftonline.com/{config.get('ckanext.msal.tenant_id', 'common')}"
REDIRECT_PATH = config.get("ckanext.msal.redirect_path", "/get_msal_token")
USER_SESSION_LIFETIME = config.get("ckanext.msal.session_lifetime", 3600)
RESTRICTED_DOMAINS = config.get("ckanext.msal.restrict.restricted_domain_list")
ALLOWED_DOMAINS = config.get("ckanext.msal.restrict.allowed_domain_list")
RESTRICTION_ERR = config.get(
    "ckanext.msal.restrict.error_message",
    "Your email domain is restricted. Please, contact site admin.",
)

SCOPE = ["User.ReadBasic.All"]



import secrets
from typing import Dict, Optional, Any, List
from datetime import datetime as dt
from datetime import timedelta as td

import msal

import ckan.lib.helpers as h
import ckan.plugins.toolkit as tk
from ckan.common import session

import ckanext.msal.config as msal_conf


def build_msal_app(cache=None, authority=None):
    return msal.ConfidentialClientApplication(
        msal_conf.CLIENT_ID,
        authority=authority or msal_conf.AUTHORITY,
        client_credential=msal_conf.CLIENT_SECRET,
        token_cache=cache,
    )


def build_auth_code_flow(authority=None, scopes=None) -> Dict[str, Any]:
    return build_msal_app(authority=authority).initiate_auth_code_flow(
        scopes or [], redirect_uri=h.url_for("msal.authorized", _external=True)
    )


def _load_cache() -> msal.SerializableTokenCache:
    cache = msal.SerializableTokenCache()
    if session.get("msal_token_cache"):
        cache.deserialize(session["msal_token_cache"])
    return cache


def _save_cache(cache) -> None:
    if cache.has_state_changed:
        session["msal_token_cache"] = cache.serialize()


def _get_token_from_cache(scope=None) -> Optional[Dict[Any, Any]]:
    cache = _load_cache()
    app = build_msal_app(cache=cache)
    accounts: List[Dict[Any, Any]] = app.get_accounts()

    if accounts:
        _save_cache(cache)
        return app.acquire_token_silent(scope, account=accounts[0])


def _get_exp_date():
    """
    Returns a float number that represents an expiration date of user session
    The session lifetime is configurable with `ckanext.msal.session_lifetime` option

    return
    type: float
    """
    return (dt.now() + td(seconds=msal_conf.USER_SESSION_LIFETIME)).timestamp()


def _make_password() -> str:
    """
    Return a random URL-safe text string, in Base64 encoding

    return
    type: str
    """
    return secrets.token_urlsafe(60)


def get_restricted_domains() -> List[str]:
    """Returns a lits of restricted domains from config
    User won't be able to login with this email domain

    Returns:
        List[str]: a list of domain strings
    """

    return tk.aslist(msal_conf.RESTRICTED_DOMAINS, ",")


def get_allowed_domains() -> List[str]:
    """Returns a lits of allowed domains from config
    User will be able to login only with those email domains

    Returns:
        List[str]: a list of domain strings
    """

    return tk.aslist(msal_conf.ALLOWED_DOMAINS, ",")


def is_email_restricted(email: str) -> bool:
    """Checks if the user email is restricted by domain
    Returns True if resticted

    Args:
        email (str): user email

    Returns:
        bool
    """

    for domain in get_restricted_domains():
        if email.endswith(domain):
            return True
    return False


def is_email_allowed(email: str) -> bool:
    """Checks if the user email is allowed by domain
    Returns True if allowed

    Args:
        email (str): user email

    Returns:
        bool
    """

    for domain in get_restricted_domains():
        if email.endswith(domain):
            return True
    return False



from datetime import datetime as dt

import pytest

from ckan.tests import factories
import ckanext.msal.user as user_funcs


@pytest.mark.ckan_config("ckan.plugins", "msal")
@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestUser(object):
    user_dict = {
        "@odata.context": "https://graph.microsoft.com/v1.0/$metadata#users/$entity",
        "@odata.id": "https://graph.microsoft.com/v2/3c8827a9-65fe-40b5-8644-3173d7026601/directoryObjects/fb9c93ba-0768-4816-8fcc-802b588fb8bf/Microsoft.DirectoryServices.User",
        "businessPhones": ["380999222611"],
        "displayName": "Mark Spencer",
        "givenName": "Mark",
        "mailNickname": "mark209",
        "jobTitle": None,
        "mail": None,
        "mobilePhone": None,
        "officeLocation": None,
        "preferredLanguage": None,
        "surname": "Spencer",
        "userPrincipalName": "kvaqich@kvaqich.onmicrosoft.com",
        "id": "fb9c93ba-0768-4816-8fcc-802b588fb8bf",
    }

    def test_get_email_upn(self):
        """
        Returns userPrincipalName from user_dict
        """
        email = user_funcs._get_email(self.user_dict)
        assert email == "kvaqich@kvaqich.onmicrosoft.com"

    def test_get_email_mail(self):
        """
        Returns `mail` field if `userPrincipalName` is empty
        """
        email = user_funcs._get_email({"mail": "kvaqich@kvaqich.onmicrosoft.com"})
        assert email == "kvaqich@kvaqich.onmicrosoft.com"

    def test_get_email_empty(self):
        """
        Returns random email if both fields are empty
        """
        email = user_funcs._get_email({})
        assert email

    def test_get_username(self):
        """
        Fetches username from user_dict
        mailNickname -> munge(mail) -> munge(userPrincipalName)
        """
        username = user_funcs._get_username(self.user_dict)
        username == "mark209"

        username = user_funcs._get_username({"mail": "testuser@gmail.com"})
        username == "testuser"

        username = user_funcs._get_username({"userPrincipalName": "testuser2@gmail.com"})
        username == "testuser2"

    def test_is_username_unqiue(self):
        user = factories.User()

        is_unique: bool = user_funcs._is_username_unique(user['name'])
        assert not is_unique

        is_unique: bool = user_funcs._is_username_unique("absolutely_new_user")
        assert is_unique
    
